@@ -1,4 +1,10 @@
-from typing import Optional, Any, List, Dict
+from typing import Optional, Any, List, Dict, ForwardRef
+
+import re
+import builtins
+import keyword
+import uuid
+
 
 from pydantic import BaseModel, Field, root_validator, Extra
 
@@ -70,29 +76,59 @@ class PathsBase(ObjectBase):
 class RootBase:
     @staticmethod
     def resolve(api, root, obj, _PathItem, _Reference):
+        from . import v20, v30, v31
+
+        def replaceSchemaReference(data):
+            def replace(value):
+                if not isinstance(value, SchemaBase):
+                    return value
+                r = getattr(value, "ref", None)
+                if not r:
+                    return value
+                return _Reference.construct(ref=r)
+
+            if isinstance(data, list):
+                for idx, item in enumerate(data):
+                    n = replace(item)
+                    if item != n:
+                        data[idx] = n
+
+            elif isinstance(data, dict):
+                new = dict()
+                for k, v in data.items():
+                    n = replace(v)  # Swagger 2.0 Schema.ref resolver …
+                    if v != n:
+                        v = n
+                        new[k] = v
+                if new:
+                    data.update(new)
+
         if isinstance(obj, ObjectBase):
-            for slot in filter(lambda x: not x.startswith("_"), obj.__fields_set__):
+            for slot in filter(lambda x: not x.startswith("_") or x == "__root__", obj.__fields_set__):
                 value = getattr(obj, slot)
                 if value is None:
                     continue
 
                 # v3.1 - Schema $ref
-                if isinstance(value, SchemaBase):
-                    r = getattr(value, "ref", None)
-                    if r is not None:
-                        value = _Reference.construct(ref=r)
-                        setattr(obj, slot, value)
+                if isinstance(root, (v20.root.Root, v30.root.Root, v31.root.Root)):
+                    if isinstance(value, SchemaBase):
+                        r = getattr(value, "ref", None)
+                        if r and not isinstance(r, ReferenceBase):
+                            value = _Reference.construct(ref=r)
+                            setattr(obj, slot, value)
 
-                """
-                ref fields embedded in objects -> replace the object with a Reference object
+                if not isinstance(value, ReferenceBase):
+                    """
+                    ref fields embedded in objects -> replace the object with a Reference object
 
-                PathItem Ref is ambigous
-                https://github.com/OAI/OpenAPI-Specification/issues/2635
-                """
-                if isinstance(obj, _PathItem) and slot == "ref":
-                    ref = _Reference.construct(ref=value)
-                    ref._target = api.resolve_jr(root, obj, ref)
-                    setattr(obj, slot, ref)
+                    PathItem Ref is ambigous
+                    https://github.com/OAI/OpenAPI-Specification/issues/2635
+                    """
+                    if isinstance(root, (v20.root.Root, v30.root.Root, v31.root.Root)):
+                        if isinstance(obj, _PathItem) and slot == "ref":
+                            ref = _Reference.construct(ref=value)
+                            ref._target = api.resolve_jr(root, obj, ref)
+                            setattr(obj, slot, ref)
 
                 value = getattr(obj, slot)
 
@@ -110,13 +146,24 @@ class RootBase:
                 else:
                     raise TypeError(type(value), value)
         elif isinstance(obj, dict):
+            if isinstance(root, (v20.root.Root, v31.root.Root)):
+                """
+                Resolving/Replacing Swagger 2.0 nested Schema.ref
+                Schema.properties[name] -> Schema.ref ==> Schema.properties[name] -> Reference
+                """
+                replaceSchemaReference(obj)
+
             for k, v in obj.items():
                 if isinstance(v, _Reference):
                     if v.ref:
                         v._target = api.resolve_jr(root, obj, v)
                 elif isinstance(v, (ObjectBase, dict, list)):
                     RootBase.resolve(api, root, v, _PathItem, _Reference)
+
         elif isinstance(obj, list):
+            if isinstance(root, (v20.root.Root, v31.root.Root)):
+                replaceSchemaReference(obj)
+
             # if it's a list, resolve its item's references
             for item in obj:
                 if isinstance(item, _Reference):
@@ -129,11 +176,7 @@ class RootBase:
         Resolves all reference objects below this object and notes their original
         value was a reference.
         """
-        # don't circular import
-
-        root = self
-
-        RootBase.resolve(api, self, self, None, None)
+        # RootBase.resolve(api, self, self, None, None)
         raise NotImplementedError("specific")
 
     def resolve_jp(self, jp):
@@ -150,7 +193,7 @@ class RootBase:
         path = jp.split("/")[1:]
         node = self
 
-        for part in path:
+        for idx, part in enumerate(path, start=1):
             part = JSONPointer.decode(part)
 
             if isinstance(node, PathsBase):  # forward
@@ -158,18 +201,17 @@ class RootBase:
 
             if isinstance(node, dict):
                 if part not in node:  # pylint: disable=unsupported-membership-test
-                    raise ReferenceResolutionError(f"Invalid path {path} in Reference")
+                    raise ReferenceResolutionError(f"Invalid path {path[:idx]} in Reference")
                 node = node.get(part)
             elif isinstance(node, list):
                 node = node[int(part)]
             elif isinstance(node, ObjectBase):
-                if part == "schema":
-                    part = "schema_"
+                part = Model.nameof(part)
                 if not hasattr(node, part):
-                    raise ReferenceResolutionError(f"Invalid path {path} in Reference")
+                    raise ReferenceResolutionError(f"Invalid path {path[:idx]} in Reference")
                 node = getattr(node, part)
             else:
-                raise TypeError(node)
+                raise ReferenceResolutionError(f"Invalid node {node} in Reference {path[:idx]}")
 
         return node
 
@@ -187,6 +229,45 @@ class DiscriminatorBase:
 
 
 class SchemaBase:
+    def __getstate__(self):
+        """
+        pickle can't do the _model_type - remove from pydantic's __getstate__
+        :return:
+        """
+        r = BaseModel.__getstate__(self)
+        try:
+            if "_model_type" in r["__private_attribute_values__"]:
+                r["__private_attribute_values__"] = r["__private_attribute_values__"].copy()
+                del r["__private_attribute_values__"]["_model_type"]
+        except Exception:
+            pass
+        return r
+
+    def _get_identity(self, prefix="XLS", name=None):
+        if not hasattr(self, "_identity"):
+            if name is None:
+                name = self.title
+            if name:
+                n = re.sub(r"[^\w]", "_", name, flags=re.ASCII)
+            else:
+                n = str(uuid.uuid4()).replace("-", "_")
+
+            try:
+                # n = re.sub(r"^([0-9]+)(.*)", r"CLS\1\2", n)
+                int(n[0])
+                n += "_"
+            except ValueError:
+                pass
+
+            if keyword.iskeyword(n) or hasattr(builtins, n):
+                n += "_"
+
+            if n != name:
+                self._identity = f"{prefix}{n}"
+            else:
+                self._identity = name
+        return self._identity
+
     def set_type(
         self, names: List[str] = None, discriminators: List[DiscriminatorBase] = None, extra: "SchemaBase" = None
     ):
@@ -199,7 +280,11 @@ class SchemaBase:
             return Model.from_schema(self, names, discriminators, extra)
 
     def get_type(
-        self, names: List[str] = None, discriminators: List[DiscriminatorBase] = None, extra: "SchemaBase" = None
+        self,
+        names: List[str] = None,
+        discriminators: List[DiscriminatorBase] = None,
+        extra: "SchemaBase" = None,
+        fwdref: bool = False,
     ):
         try:
             if extra is None:
@@ -207,7 +292,14 @@ class SchemaBase:
             else:
                 return self.set_type(names, discriminators, extra)
         except AttributeError:
-            return self.set_type(names, discriminators)
+            if fwdref:
+                if "module" in ForwardRef.__init__.__code__.co_varnames:
+                    # FIXME Python < 3.9 compat
+                    return ForwardRef(self._get_identity("FWD"), module="aiopenapi3.me")
+                else:
+                    return ForwardRef(self._get_identity("FWD"))
+            else:
+                return self.set_type(names, discriminators)
 
     def model(self, data: Dict):
         """
@@ -230,3 +322,6 @@ class SchemaBase:
             return [self.items.get_type().parse_obj(i) for i in data]
         else:
             return self.get_type().parse_obj(data)
+
+
+from .model import Model
